#include <xmms/plugin.h>
#include <xmms/util.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include        "Types.h"
#include        "SPC700.h"
#include        "DSP.h"
#include        "APU.h"
#include	"spcload.h"
#include	"config.h"

static volatile int RequestSeek = -1;
static volatile int RequestStop = 0;
static volatile int SongEnded = 0;

static pthread_mutex_t playmutex = PTHREAD_MUTEX_INITIALIZER;

static SPCINFO SPCInfo;

extern InputPlugin sexy_ip;

InputPlugin *get_iplugin_info(void)
      {
         sexy_ip.description = "sexySPC SPC Player 0.2.1";
         return &sexy_ip;
      }

static void *pAPURAM;

void init(void)
{
  //Allocate RAM -----------------------------
  void *pAPUBase = malloc(131072);

  pAPURAM = (void*)(((u32)pAPUBase + 0xFFFF) & ~0xFFFF);

  InitAPU(pAPURAM);
}

static void about(void)
{

}

static void config(void)
{

}

static int testfile(char *fn)
{
 FILE *fp;
 char buf[0x1b+1];
 char *testie="SNES-SPC700 Sound File Data";
 
 if(!strncasecmp(fn,"http://",7)) return(0); // Don't handle http://blahblah
 if(!(fp=fopen(fn,"rb"))) return(0);
 if(fread(buf,1,0x1b,fp)!=0x1b) {fclose(fp);return(0);}
 fclose(fp);

 if(strncmp(buf,testie,0x1b)) return(0);
 return(1);
}

static void SI(void)
{
 char tmp[1024];
 if(SPCInfo.song[0])
  sprintf(tmp,"%s - %s",SPCInfo.game,SPCInfo.song);
 else
  strcpy(tmp,SPCInfo.game);
 sexy_ip.set_info(tmp,SPCInfo.totallength/64,SexyConfig.rate*SexyConfig.channels*abs(SexyConfig.bits),SexyConfig.rate,SexyConfig.channels);
}

int ToXMMSBits(void)
{
 switch(abs(SexyConfig.bits))
 {
  case 8: return(FMT_U8);
  case 16: return(FMT_S16_LE);
 }
 return(0);
}

static pthread_t dethread;
void sexyd_update(void)
{
 char buffer[512*4];
 char *Buffer=buffer;
 int count=512*4;
 int mask;

 EmuAPU(Buffer, 512, 1);
 sexy_ip.add_vis_pcm(sexy_ip.output->written_time(), ToXMMSBits(), SexyConfig.channels, count/4, Buffer);

 mask = ~((((SexyConfig.bits / 8) * SexyConfig.channels)) - 1);
 //printf("%d\n",mask&255);
 
 while(count>0 && !RequestStop)
 {
  int t=sexy_ip.output->buffer_free() & mask;
  if(t>count)
  {
   sexy_ip.output->write_audio(Buffer,count);
  }
  else
  {
   unsigned long long tmp;
   if(t)
    sexy_ip.output->write_audio(Buffer,t);

   tmp=count-t;
   tmp*=1000*1000;	// Multiply so we'll have microseconds after the following divide.
   tmp/=SexyConfig.rate*SexyConfig.channels*abs(SexyConfig.bits);
   usleep(tmp);
  }
  count-=t;
  Buffer+=t;
 }
}

static void SeekWrapper(long t)	// BLAH, SNESAPU crashes on SeekAPU.  Why?
{
 char Buffer[2048*4];

 t /= 2;

 while(t > 0)
 {
  if(t>=2048)
   EmuAPU(Buffer, 2048, 1);
  else EmuAPU(Buffer, t, 1);
  t-= 2048;
 }
}

static void *playloop(void *arg)
{
 static long inseek;
 inseek=0;

 for(;;)
 {
  pthread_mutex_lock(&playmutex);

  if(RequestStop) 
   break;

  if(t64Cnt>=SPCInfo.totallength && SPCInfo.totallength) 
  {
   if(!sexy_ip.output->buffer_playing())
    break;
   if(!inseek) usleep(1000);
  }
  else if(!inseek)
   sexyd_update();

  if(RequestSeek != -1)
  {
   unsigned long t=RequestSeek;

   inseek=t*64000-t64Cnt;

   if(inseek<0)	/* Reverse seek requested. */
   {
    SetSPC(&SPCInfo, pAPURAM);
    inseek=t*64000;
    sexy_ip.output->flush(0);
   }
   RequestSeek = -1;
  }

  if(inseek)
  {
   if(inseek<64000)
   {
    SeekWrapper(inseek);
    //SeekAPU(inseek,0);
    inseek=0;
    sexy_ip.output->flush(t64Cnt / 64);
   }
   else
   {
    SeekWrapper(64000);
    //SeekAPU(64000,0);
    inseek-=64000;
    sexy_ip.output->flush(t64Cnt / 64);
   }
  }
  pthread_mutex_unlock(&playmutex);
 }

 sexy_ip.output->close_audio();
 SongEnded = 1;
 pthread_mutex_unlock(&playmutex);

 pthread_exit(0);
 return(0);
}

static int paused;

void play(char *fn)
{
 if(!sexy_ip.output->open_audio(ToXMMSBits(), SexyConfig.rate, SexyConfig.channels))
 {
  puts("Error opening audio.");
  return;
 }

 paused = 0;

 if(LoadSPC(fn,&SPCInfo,pAPURAM))
 {
  RequestSeek = -1;
  RequestStop = 0;
  SongEnded = 0;
  SI();

  if(pthread_mutex_init(&playmutex, NULL))
   puts("Mutex failure!");
  pthread_create(&dethread,0,playloop,0);
 }
}

void stop(void)
{
 RequestStop = 1;
 //puts("Stopping");
 pthread_mutex_lock(&playmutex);
 //puts("Locked");
 if(!SongEnded)
 {
  if(paused)
   sexy_ip.output->pause(0);
  paused = 0;
 }
 pthread_mutex_unlock(&playmutex);

 pthread_join(dethread,0);
 pthread_mutex_destroy(&playmutex);
 KillSPC(&SPCInfo);
}

void sexy_pause(short p)
{
 sexy_ip.output->pause(p);
 paused = p;
}

void seek(int time)
{
 pthread_mutex_lock(&playmutex);
 RequestSeek = time;
 pthread_mutex_unlock(&playmutex);
}

int gettime(void)
{
 if(SongEnded && !RequestStop)
  return(-1);
 else
  return sexy_ip.output->output_time();
}

void getsonginfo(char *fn, char **title, int *length)
{
  SPCINFO tmp;

  if(LoadSPC(fn,&tmp,0))
  {
   *length=tmp.totallength/64;
   if(tmp.song[0])
   {
    *title=malloc(strlen(tmp.game)+strlen(tmp.song)+3+1);
    sprintf(*title,"%s - %s",tmp.game,tmp.song);
   }
   else
   {
    *title=malloc(strlen(tmp.game)+1);
    strcpy(*title,tmp.game);
   }
  }
}

InputPlugin sexy_ip =
{
 0,0,"SPC Player",init,about,config,testfile,0,play,stop,sexy_pause,
 seek,0,gettime,0,0,0,0,0,0,0,getsonginfo,0,0
};

#ifdef MOOO
/* For debugging */
int main(int argc, char *argv[])
{
  void *pAPUBase = malloc(131072);

  pAPURAM = (void*)(((u32)pAPUBase + 0xFFFF) & ~0xFFFF);

  InitAPU(pAPURAM);

  if(LoadSPC(argv[1],&SPCInfo,pAPURAM))
  {
   printf("%s\n",SPCInfo.game);
   SetSPC(&SPCInfo, pAPURAM);

   for(;;)
   {
    char Buffer[512 * 8];

    printf("Buffer: %08x:%08x\n",Buffer,&Buffer[512 * 8]);
    puts("Beginning...");
    EmuAPU(Buffer, 512, 1);
    puts("Done");

   }
  }

 return(0);
}
#endif
